/**
 * The MIT License (MIT)
 *
 * Copyright (c) 2021 Marcus Britanicus (https://gitlab.com/marcusbritanicus)
 * Copyright (c) 2021 Abrar (https://gitlab.com/s96Abrar)
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 **/

#include <stdlib.h>
#include <unistd.h>

#include <wayqt/Application.hpp>
#include <QWidget>

#include <wayqt/WlGlobal.hpp>
#include <wayqt/Registry.hpp>

static inline QString getSocketName( QString appId ) {
    /** Get the env-var XDG_RUNTIME_DIR */
    QString sockName( qgetenv( "XDG_RUNTIME_DIR" ) );

    /** The env-var was not set. We will use /tmp/ */
    if ( not sockName.count() ) {
        sockName = "/tmp/";
    }

    if ( not sockName.endsWith( "/" ) ) {
        sockName += "/";
    }

    /** Append a random number */
    sockName += QString( "%1-Scoket-%2" ).arg( appId ).arg( getuid() );

    return sockName;
}


WQt::Application::Application( const QString& appId, int& argc, char **argv ) : QApplication( argc, argv ) {
    /* App ID */
    mAppId = appId;

    /* Socket Name */
    mSocketName = getSocketName( appId );

    wl_display *display = WQt::getWlDisplay();

    if ( not display ) {
        qDebug() << "Unable to acquire wl_display from the compositor.";
        qDebug() << "Your experience will be severly limited.";

        registry = nullptr;
    }

    else {
        registry = new WQt::Registry( display );
        QObject::connect(
            registry, &WQt::Registry::errorOccured, [ = ] ( WQt::Registry::ErrorType et ) {
                qDebug() << "Error caused on registry" << et;
                qDebug() << "Valiantly trying to continue...";
            }
        );

        registry->setup();
    }

    /* Lock File */
    lockFile = new QLockFile( mSocketName + ".lock" );

    /* Try to lock the @lockFile, if it fails, then we're not the first instance */
    if ( lockFile->tryLock() ) {
        /* Local Server for communication */
        mServer = new QLocalServer( this );

        /* Start the server */
        bool res = mServer->listen( mSocketName );

        /*
         * @res can't be false at the moment, because we're the first instance.
         * The only reason why @res is false, the socket file exists from a previous
         * crash. So delete it and try again.
         */
        if ( not res && (mServer->serverError() == QAbstractSocket::AddressInUseError) ) {
            QLocalServer::removeServer( mSocketName );
            res = mServer->listen( mSocketName );

            if ( !res ) {
                qWarning( "WQt::Application: listen on local socket failed, %s", qPrintable( mServer->errorString() ) );
            }
        }

        /* Irrespective of what happens, we will try to connect newConnection to receiveConnection */
        QObject::connect( mServer, &QLocalServer::newConnection, this, &WQt::Application::handleConnection );
    }
}


WQt::Application::~Application() {
    disconnect();

    if ( mServer ) {
        mServer->deleteLater();
    }

    delete lockFile;
}


bool WQt::Application::isRunning() {
    /*
     * If we have the lock, we're the server
     * In other words, if we're not there, there is no server
     */
    if ( lockFile->isLocked() ) {
        return false;
    }

    /* If we cannot get the lock then the server is running elsewhere */
    if ( not lockFile->tryLock() ) {
        return true;
    }

    /* Be default, we'll assume that the server is running elsewhere */
    return true;
}


QString WQt::Application::id() const {
    return mAppId;
}


bool WQt::Application::sendMessage( const QString& message ) {
    if ( not isRunning() ) {
        return false;
    }

    /* Preparing socket */
    QLocalSocket socket( this );

    /* Connecting to server */
    socket.connectToServer( mSocketName );

    /* Wait for ACK (500 ms) */
    if ( not socket.waitForConnected( 500 ) ) {
        return false;
    }

    /* Send the message to the server */
    socket.write( message.toUtf8() );

    /** Should finish writing in 500 ms */
    return socket.waitForBytesWritten( 500 );
}


void WQt::Application::disconnect() {
    if ( mServer ) {
        mServer->close();
    }

    lockFile->unlock();
}


void WQt::Application::setActivationWindow( QWidget *aw, bool activateOnMessage ) {
    /* Activation makes sense only when the server is active */
    if ( mServer ) {
        actWin = aw;

        if ( activateOnMessage ) {
            QObject::connect( this, &WQt::Application::messageReceived, this, &WQt::Application::activateWindow );
        }

        else {
            QObject::disconnect( this, &WQt::Application::messageReceived, this, &WQt::Application::activateWindow );
        }
    }
}


QWidget *WQt::Application::activationWindow() const {
    return actWin;
}


void WQt::Application::activateWindow() {
    if ( actWin ) {
        actWin->show();
        actWin->setWindowState( actWin->windowState() & ~Qt::WindowMinimized );
        actWin->raise();
        actWin->activateWindow();
    }
}


void WQt::Application::handleConnection() {
    /* Preparing socket */
    QLocalSocket *socket = mServer->nextPendingConnection();

    if ( not socket ) {
        return;
    }

    socket->waitForReadyRead( 500 );
    QByteArray msg = socket->readAll();

    /** Close the connection */
    socket->close();

    emit messageReceived( QString( msg ) );
}


WQt::Registry *WQt::Application::waylandRegistry() {
    return registry;
}
