/**
 * The MIT License (MIT)
 *
 * Copyright (c) 2021 Marcus Britanicus (https://gitlab.com/marcusbritanicus)
 * Copyright (c) 2021 Abrar (https://gitlab.com/s96Abrar)
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 **/

#include <QGuiApplication>
#include <QWindow>
#include <QScreen>

#include <qpa/qplatformnativeinterface.h>
#include <wayland-client.h>

#include "wayqt/WlGlobal.hpp"

wl_display *WQt::getWlDisplay() {
    QPlatformNativeInterface *native = QGuiApplication::platformNativeInterface();

    if ( !native ) {
        return nullptr;
    }

    struct wl_display *display = reinterpret_cast<wl_display *>(native->nativeResourceForWindow( "display", nullptr ) );

    return display;
}


wl_output *WQt::getWlOutput( QScreen *screen ) {
    QPlatformNativeInterface *native = QGuiApplication::platformNativeInterface();

    if ( !native ) {
        return nullptr;
    }

    struct wl_output *output = reinterpret_cast<wl_output *>(native->nativeResourceForScreen( "output", screen ) );

    return output;
}


wl_surface *WQt::getWlSurface( QWindow *window ) {
    window->create();
    QPlatformNativeInterface *native = QGuiApplication::platformNativeInterface();

    if ( !native ) {
        return nullptr;
    }

    struct wl_surface *surface = reinterpret_cast<wl_surface *>(native->nativeResourceForWindow( "surface", window ) );

    return surface;
}


bool WQt::isRunning() {
    /* Check if XDG_SESSION_TYPE is set */
    QString session = qgetenv( "XDG_SESSION_TYPE" );

    if ( session.toLower() == QStringLiteral( "wayland" ) ) {
        return true;
    }

    /*
     * May be XDG_SESSION_TYPE is not set. Try Harder.
     * We check if WAYLAND_DISPLAY is set.
     */
    QString wlID = qgetenv( "WAYLAND_DISPLAY" );

    if ( not wlID.isEmpty() ) {
        return true;
    }

    /*
     * May be WAYLAND_DISPLAY is not set. Try Harder still.
     * We check if DESKTOP_SESSION is set.
     */
    QString deSess = qgetenv( "DESKTOP_SESSION" );

    if ( deSess.contains( "wayland-sessions" ) ) {
        return true;
    }

    /* It's probably not a wayland session. */
    return false;
}


bool WQt::isWayfire() {
    /** Checking the WAYFIRE_CONFIG_FILE variable is sufficient for our cause */
    QString wfConfigFile = qgetenv( "WAYFIRE_CONFIG_FILE" );

    if ( wfConfigFile.count() ) {
        return true;
    }

    return false;
}
